/*
Package config reads the server configuration from disk.
*/
package config

import (
	"encoding/json"
	"fmt"
	"io/ioutil"
	"log"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"time"
)

var (
	version = Version{
		Major: 0,
		Minor: 1,
		Patch: 0,
	}

	defaultConfig = Settings{
		Version: version,
		Socket:  &defaultSocket,
		Engines: &defaultEngines,
	}

	defaultSocket = Socket{
		StdIO: &defaultStdIO,
	}

	defaultStdIO = StdIO{}

	defaultWebsocket = Websocket{
		Port: 15291,
	}

	defaultSecureWebsocket = SecureWebsocket{
		Port:        15292,
		Certificate: "cert.pem",
		PrivateKey:  "privkey.pem",
	}

	defaultRateLimit = RateLimit{
		Points: &defaultPoints,
	}

	defaultPoints = Points{
		Grace:   1.05,
		Max:     3000,
		Tick:    1,
		Timeout: 500,
	}

	defaultEngines = Engines{
		Path:    "engines-enabled",
		Engines: NamedEngines{},
	}
)

func init() {
	var err error
	if defaultRateLimit.Period, err = time.ParseDuration("30m"); err != nil {
		log.Println("config:", err)
	}
	if defaultRateLimit.Tick, err = time.ParseDuration("10ms"); err != nil {
		log.Println("config:", err)
	}
	if defaultRateLimit.Transition, err = time.ParseDuration("1m"); err != nil {
		log.Println("config:", err)
	}
	if defaultRateLimit.Timeout, err = time.ParseDuration("5s"); err != nil {
		log.Println("config:", err)
	}
}

// FieldRequiredError indicates a missing field in the configuration file.
type FieldRequiredError struct {
	Field string
}

// Settings stores the configuration state of the server.
type Settings struct {
	File    string   `json:"-"`
	Version Version  `json:"-"`
	Socket  *Socket  `json:"socket"`
	Engines *Engines `json:"engines"`
}

// Version stores the version information of the server.
type Version struct {
	Major uint `json:"major"`
	Minor uint `json:"minor"`
	Patch uint `json:"patch"`
}

// Socket is one of the following: standard I/O, websocket or secure websocket
type Socket struct {
	StdIO           *StdIO           `json:"stdio"`
	Websocket       *Websocket       `json:"websocket,omitempty"`
	SecureWebsocket *SecureWebsocket `json:"secure-websocket,omitempty"`
}

// StdIO marks the use of standard I/O streams.
type StdIO struct{}

// Websocket holds configurations options concerning the standard websocket use.
type Websocket struct {
	MaxConnections uint       `json:"max-connections"`
	Port           uint       `json:"port"`
	RateLimit      *RateLimit `json:"rate-limit"`
}

// SecureWebsocket holds configuration options concerning the secure websocket use.
type SecureWebsocket struct {
	MaxConnections uint       `json:"max-connections"`
	Port           uint       `json:"port"`
	RateLimit      *RateLimit `json:"rate-limit"`
	Certificate    string     `json:"certificate"`
	PrivateKey     string     `json:"private-key"`
}

// RateLimit configures the limits for abusive connections.
type RateLimit struct {
	Period     time.Duration `json:"period"`
	Tick       time.Duration `json:"tick"`
	Transition time.Duration `json:"transition"`
	Timeout    time.Duration `json:"timeout"`
	Points     *Points       `json:"points"`
}

// Points configures the penalties for abusive connections.
type Points struct {
	Grace   float64 `json:"grace"`
	Max     uint    `json:"max"`
	Tick    uint    `json:"tick"`
	Timeout uint    `json:"timeout"`
}

// Engines holds the configuration path and the configuration of analysis engines.
type Engines struct {
	Path    string       `json:"path"`
	Engines NamedEngines `json:"-"`
}

// NamedEngines is a shorthand for ID - Engine association.
type NamedEngines map[string]Engine

// Engine represents a single analysis engine configuration.
type Engine struct {
	Name        *string  `json:"name"`
	Version     *Version `json:"version"`
	Author      string   `json:"author"`
	Description string   `json:"description"`
	Url         string   `json:"url"`
	Command     *string  `json:"command"`
	Arguments   []string `json:"arguments"`
}

func (e *FieldRequiredError) Error() string {
	return "load config: field missing: " + e.Field
}

func newSettings(other *Settings) (s *Settings) {
	if other == nil {
		return
	}

	return &Settings{
		File:    other.File,
		Version: other.Version,
		Socket:  newSocket(other.Socket),
		Engines: newEngines(other.Engines),
	}
}

func (s *Settings) UnmarshalJSON(data []byte) (err error) {
	type settings Settings
	var raw settings

	err = json.Unmarshal(data, &raw)

	if err != nil {
		return
	}
	if raw.Socket == nil {
		raw.Socket = newSocket(&defaultSocket)
	}
	if raw.Engines == nil {
		raw.Engines = newEngines(&defaultEngines)
	}

	s.Socket = raw.Socket
	s.Engines = raw.Engines

	return
}

func (v Version) String() string {
	return fmt.Sprintf("v%d.%d.%d", v.Major, v.Minor, v.Patch)
}

func newSocket(other *Socket) (s *Socket) {
	if other == nil {
		return
	}

	s = &Socket{}
	if other.StdIO != nil {
		s.StdIO = &StdIO{}
	}
	if other.Websocket != nil {
		ws := *other.Websocket
		s.Websocket = &ws
	}
	if other.SecureWebsocket != nil {
		ws := *other.SecureWebsocket
		s.SecureWebsocket = &ws
	}

	return
}

func (s *Socket) UnmarshalJSON(data []byte) (err error) {
	type socket Socket
	var raw socket

	err = json.Unmarshal(data, &raw)

	if err != nil {
		return
	}
	if raw.StdIO == nil && raw.Websocket == nil && raw.SecureWebsocket == nil {
		//raw = socket(*newSocket(&defaultSocket))
		return
	}

	*s = Socket(raw)

	return
}

func (ws *Websocket) UnmarshalJSON(data []byte) (err error) {
	type websocket Websocket
	var raw websocket

	err = json.Unmarshal(data, &raw)

	if err != nil {
		return
	}
	if raw.Port == 0 {
		raw.Port = defaultWebsocket.Port
	}

	*ws = Websocket(raw)

	return
}

func (ws *SecureWebsocket) UnmarshalJSON(data []byte) (err error) {
	type secureWebsocket SecureWebsocket
	var raw secureWebsocket

	err = json.Unmarshal(data, &raw)

	if err != nil {
		return
	}
	if raw.Port == 0 {
		raw.Port = defaultSecureWebsocket.Port
	}
	if raw.Certificate == "" {
		raw.Certificate = defaultSecureWebsocket.Certificate
	}
	if raw.PrivateKey == "" {
		raw.PrivateKey = defaultSecureWebsocket.PrivateKey
	}

	*ws = SecureWebsocket(raw)

	return
}

func (rl *RateLimit) UnmarshalJSON(data []byte) (err error) {
	type rateLimit struct {
		Period     string  `json:"period"`
		Tick       string  `json:"tick"`
		Transition string  `json:"transition"`
		Timeout    string  `json:"timeout"`
		Points     *Points `json:"points"`
	}
	var raw rateLimit

	err = json.Unmarshal(data, &raw)

	if err != nil {
		return
	}
	if raw.Points == nil {
		points := defaultPoints
		raw.Points = &points
	}
	if raw.Period == "" {
		raw.Period = defaultRateLimit.Period.String()
	}
	if raw.Tick == "" {
		raw.Tick = defaultRateLimit.Tick.String()
	}
	if raw.Transition == "" {
		raw.Transition = defaultRateLimit.Transition.String()
	}
	if raw.Timeout == "" {
		raw.Timeout = defaultRateLimit.Timeout.String()
	}

	period, err := time.ParseDuration(raw.Period)
	if err != nil {
		log.Println("config:", err)
	}
	tick, err := time.ParseDuration(raw.Tick)
	if err != nil {
		log.Println("config:", err)
	}
	transition, err := time.ParseDuration(raw.Transition)
	if err != nil {
		log.Println("config:", err)
	}
	timeout, err := time.ParseDuration(raw.Timeout)
	if err != nil {
		log.Println("config:", err)
	}

	*rl = RateLimit{
		Period:     period,
		Tick:       tick,
		Transition: transition,
		Timeout:    timeout,
		Points:     raw.Points,
	}

	return
}

func (p *Points) UnmarshalJSON(data []byte) (err error) {
	type points Points
	var raw points

	err = json.Unmarshal(data, &raw)

	if err != nil {
		return
	}
	if raw.Grace < 1.0 {
		raw.Grace = defaultPoints.Grace
	}
	if raw.Max == 0 {
		raw.Max = defaultPoints.Max
	}
	if raw.Tick == 0 {
		raw.Tick = defaultPoints.Tick
	}
	if raw.Timeout == 0 {
		raw.Timeout = defaultPoints.Timeout
	}

	*p = Points(raw)

	return
}

func newEngines(other *Engines) (e *Engines) {
	if other == nil {
		return
	}

	e = &Engines{
		Path:    other.Path,
		Engines: NamedEngines{},
	}
	for k, v := range other.Engines {
		e.Engines[k] = v
	}

	return
}

func (e *Engine) UnmarshalJSON(data []byte) (err error) {
	type engine Engine
	var raw engine

	err = json.Unmarshal(data, &raw)

	if err != nil {
		return
	}
	if raw.Name == nil {
		err = &FieldRequiredError{Field: ""}
		return
	}
	if raw.Version == nil {
		err = &FieldRequiredError{Field: ""}
		return
	}
	if raw.Command == nil {
		err = &FieldRequiredError{Field: ""}
		return
	}

	*e = Engine(raw)

	return
}

// Read loads the server configuration from disk.
// If the configuration file cannot be found, Read writes to disk and returns the standard configuration.
// Any error encountered while reading or creating the configuration file is returned immediately.
func Read(name string) (*Settings, error) {
	s := newSettings(&defaultConfig)
	s.File = name

	content, err := ioutil.ReadFile(name)
	if err != nil {
		if _, err := os.Stat(name); os.IsNotExist(err) {
			return create(s)
		}
		return nil, err
	}

	if err = s.parse(content); err != nil {
		return nil, err
	}

	files, err := ioutil.ReadDir(s.Engines.Path)
	if err != nil && !os.IsNotExist(err) {
		return nil, err
	}

	for _, file := range files {
		if file.IsDir() {
			continue
		}

		name := filepath.Join(s.Engines.Path, file.Name())
		content, err := ioutil.ReadFile(name)
		if err != nil {
			return nil, err
		}

		e := Engine{}
		if err = e.parse(content); err != nil {
			return nil, err
		}
		ne := s.Engines.Engines
		ne[ne.unique(*e.Name)] = e
	}

	return s, nil
}

func create(s *Settings) (*Settings, error) {
	bytes, err := json.MarshalIndent(s, "", "\t")
	bytes = append(bytes, '\n')

	if err = ioutil.WriteFile(s.File, bytes, 0644); err != nil {
		return nil, err
	}

	return s, nil
}

func (s *Settings) parse(content []byte) (err error) {
	return json.Unmarshal(content, s)
}

func (ne NamedEngines) unique(name string) string {
	parts := strings.Split(name, " ")
	if len(parts) > 1 {
		for i := range parts {
			parts[i] = string([]rune(parts[i])[0])
		}
	}
	id := strings.Join(parts, "")

	for _, ok := ne[id]; ok; _, ok = ne[id] {
		parts := strings.Split(id, "-")
		if len(parts) == 1 {
			parts = append(parts, "0")
		}
		n, err := strconv.Atoi(parts[1])
		if err != nil {
			log.Println("cannot parse integer")
		}
		parts[1] = strconv.Itoa(n + 1)
		id = strings.Join(parts, "-")
	}

	return id
}

func (e *Engine) parse(content []byte) (err error) {
	return json.Unmarshal(content, e)
}
