package config

import (
	"fmt"
	"testing"
)

func mustParse(t *testing.T, s *Settings, content []byte) {
	err := s.parse(content)
	if err != nil {
		t.Fatalf("parse() returned error %#v; want none", err)
	}
}

func (s *Settings) equal(o *Settings) (bool, string) {
	if ok, reason := equalPointer(s, o); !ok {
		return false, fmt.Sprintf("Settings: %v", reason)
	}
	if s == nil {
		return true, ""
	}

	if s.File != o.File {
		return false, fmt.Sprintf("Settings.File: have %v; want %v", o.File, s.File)
	}

	if ok, reason := s.Version.equal(o.Version); !ok {
		return false, fmt.Sprintf("Settings.%v", reason)
	}

	if ok, reason := equalPointer(s.Socket, o.Socket); !ok {
		return false, fmt.Sprintf("Settings.Socket: %v", reason)
	}
	if s.Socket != nil {
		if ok, reason := s.Socket.equal(o.Socket); !ok {
			return false, fmt.Sprintf("Settings.%v", reason)
		}
	}

	if ok, reason := equalPointer(s.Engines, o.Engines); !ok {
		return false, fmt.Sprintf("Settings.Engines: %v", reason)
	}
	if s.Engines != nil {
		if ok, reason := s.Engines.equal(o.Engines); !ok {
			return false, fmt.Sprintf("Settings.%v", reason)
		}
	}

	return true, ""
}

func equalPointer(a, b interface{}) (bool, string) {
	if a == nil && b == nil {
		return true, ""
	}
	if a == nil || b == nil {
		return false, fmt.Sprintf("nil and non-nil pointer: have %v; want %v", b, a)
	}

	return true, ""
}

func (v Version) equal(o Version) (bool, string) {
	if v.Major != o.Major {
		return false, fmt.Sprintf("Version.Major: have %v; want %v", o.Major, v.Major)
	}
	if v.Minor != o.Minor {
		return false, fmt.Sprintf("Version.Minor: have %v; want %v", o.Minor, v.Minor)
	}
	if v.Patch != o.Patch {
		return false, fmt.Sprintf("Version.Patch: have %v; want %v", o.Patch, v.Patch)
	}

	return true, ""
}

func (s *Socket) equal(o *Socket) (bool, string) {
	if ok, reason := equalPointer(s, o); !ok {
		return false, fmt.Sprintf("Socket: %v", reason)
	}
	if s == nil {
		return true, ""
	}

	if ok, reason := equalPointer(s.StdIO, o.StdIO); !ok {
		return false, fmt.Sprintf("Socket.StdIO: %v", reason)
	}
	if s.StdIO != nil {
		if ok, reason := s.StdIO.equal(o.StdIO); !ok {
			return false, fmt.Sprintf("Socket.%v", reason)
		}
	}

	if ok, reason := equalPointer(s.Websocket, o.Websocket); !ok {
		return false, fmt.Sprintf("Socket.Websocket: %v", reason)
	}
	if s.Websocket != nil {
		if ok, reason := s.Websocket.equal(o.Websocket); !ok {
			return false, fmt.Sprintf("Socket.%v", reason)
		}
	}

	if ok, reason := equalPointer(s.SecureWebsocket, o.SecureWebsocket); !ok {
		return false, fmt.Sprintf("Socket.SecureWebsocket: %v", reason)
	}
	if s.SecureWebsocket != nil {
		if ok, reason := s.SecureWebsocket.equal(o.SecureWebsocket); !ok {
			return false, fmt.Sprintf("Socket.%v", reason)
		}
	}

	return true, ""
}

func (s *StdIO) equal(o *StdIO) (bool, string) {
	if ok, reason := equalPointer(s, o); !ok {
		return false, fmt.Sprintf("StdIO: %v", reason)
	}

	return true, ""
}

func (ws *Websocket) equal(o *Websocket) (bool, string) {
	if ok, reason := equalPointer(ws, o); !ok {
		return false, fmt.Sprintf("Websocket: %v", reason)
	}
	if ws == nil {
		return true, ""
	}

	if ws.MaxConnections != o.MaxConnections {
		return false, fmt.Sprintf(
			"Websocket.MaxConnections: have %v; want %v",
			o.MaxConnections,
			ws.MaxConnections)
	}

	if ws.Port != o.Port {
		return false, fmt.Sprintf("Websocket.Port: have %v; want %v", o.Port, ws.Port)
	}

	return true, ""
}

func (ws *SecureWebsocket) equal(o *SecureWebsocket) (bool, string) {
	if ok, reason := equalPointer(ws, o); !ok {
		return false, fmt.Sprintf("SecureWebsocket: %v", reason)
	}
	if ws == nil {
		return true, ""
	}

	if ws.MaxConnections != o.MaxConnections {
		return false, fmt.Sprintf(
			"SecureWebsocket.MaxConnections: have %v; want %v",
			o.MaxConnections,
			ws.MaxConnections)
	}

	if ws.Port != o.Port {
		return false, fmt.Sprintf("SecureWebsocket.Port: have %v; want %v", o.Port, ws.Port)
	}

	if ws.Certificate != o.Certificate {
		return false, fmt.Sprintf("SecureWebsocket.Certificate: have %v; want %v", o.Certificate, ws.Certificate)
	}

	if ws.PrivateKey != o.PrivateKey {
		return false, fmt.Sprintf("SecureWebsocket.PrivateKey: have %v; want %v", o.PrivateKey, ws.PrivateKey)
	}

	return true, ""
}

func (e *Engines) equal(o *Engines) (bool, string) {
	if ok, reason := equalPointer(e, o); !ok {
		return false, fmt.Sprintf("Engines: %v", reason)
	}
	if e == nil {
		return true, ""
	}

	if e.Path != o.Path {
		return false, fmt.Sprintf("Engines.Path: have %v; want %v", o.Path, e.Path)
	}

	if len(e.Engines) != len(o.Engines) {
		return false, fmt.Sprintf("Engines.Engines.len: have %v; want %v", len(o.Engines), len(e.Engines))
	}
	for k := range e.Engines {
		if ok, reason := e.Engines[k].equal(o.Engines[k]); !ok {
			return false, fmt.Sprintf("Engines[%v].%v", k, reason)
		}
	}

	return true, ""
}

func (e Engine) equal(o Engine) (bool, string) {
	// TODO

	return true, ""
}

func TestEmptyConfiguration(t *testing.T) {
	emptyIn := []byte(`{}`)
	s := newSettings(&defaultConfig)
	mustParse(t, s, emptyIn)

	if ok, reason := defaultConfig.equal(s); !ok {
		t.Error(reason)
	}
}

func TestDefaultConfiguration(t *testing.T) {
	defaultIn := []byte((`{
	"socket": {
		"stdio": {},
		"engines": {
			"path": "engines-enabled"
		}
	}
}`))
	s := newSettings(&defaultConfig)
	mustParse(t, s, defaultIn)

	if ok, reason := defaultConfig.equal(s); !ok {
		t.Error(reason)
	}
}

func TestValidSocket(t *testing.T) {
	tests := []struct {
		name string
		in   []byte
		want *Settings
	}{
		{
			"Empty",
			[]byte(`{"socket":{}}`),
			&Settings{Socket: &defaultSocket},
		},
		{
			"StdIO",
			[]byte(`{"socket":{"stdio":{}}}`),
			&Settings{Socket: &defaultSocket},
		},
		{
			"WebsocketDefault",
			[]byte(`{"socket":{"websocket":{}}}`),
			&Settings{Socket: &Socket{Websocket: &defaultWebsocket}},
		},
		{
			"WebsocketCapped",
			[]byte(`{"socket":{"websocket":{"max-connections":2}}}`),
			&Settings{Socket: &Socket{Websocket: &Websocket{MaxConnections: 2, Port: defaultWebsocket.Port}}},
		},
		{
			"WebsocketPort",
			[]byte(`{"socket":{"websocket":{"port":4242}}}`),
			&Settings{Socket: &Socket{Websocket: &Websocket{Port: 4242}}},
		},
		{
			"SecureWebsocketDefault",
			[]byte(`{"socket":{"secure-websocket":{}}}`),
			&Settings{Socket: &Socket{SecureWebsocket: &defaultSecureWebsocket}},
		},
		{
			"SecureWebsocketCapped",
			[]byte(`{"socket":{"secure-websocket":{"max-connections":2}}}`),
			&Settings{Socket: &Socket{SecureWebsocket: &SecureWebsocket{
				MaxConnections: 2,
				Port:           defaultSecureWebsocket.Port,
				Certificate:    defaultSecureWebsocket.Certificate,
				PrivateKey:     defaultSecureWebsocket.PrivateKey,
			}}},
		},
		{
			"SecureWebsocketPort",
			[]byte(`{"socket":{"secure-websocket":{"port":4242}}}`),
			&Settings{Socket: &Socket{SecureWebsocket: &SecureWebsocket{
				Port:        4242,
				Certificate: defaultSecureWebsocket.Certificate,
				PrivateKey:  defaultSecureWebsocket.PrivateKey,
			}}},
		},
		{
			"SecureWebsocketCertificate",
			[]byte(`{"socket":{"secure-websocket":{"certificate":"test.pem"}}}`),
			&Settings{Socket: &Socket{SecureWebsocket: &SecureWebsocket{
				Port:        defaultSecureWebsocket.Port,
				Certificate: "test.pem",
				PrivateKey:  defaultSecureWebsocket.PrivateKey,
			}}},
		},
		{
			"SecureWebsocketPrivateKey",
			[]byte(`{"socket":{"secure-websocket":{"private-key":"test.pem"}}}`),
			&Settings{Socket: &Socket{SecureWebsocket: &SecureWebsocket{
				Port:        defaultSecureWebsocket.Port,
				Certificate: defaultSecureWebsocket.Certificate,
				PrivateKey:  "test.pem",
			}}},
		},
	}

	for _, test := range tests {
		t.Run(test.name, func(t *testing.T) {
			s := &Settings{}
			mustParse(t, s, test.in)
			if ok, reason := test.want.equal(s); !ok {
				t.Error(reason)
			}
		})
	}
}

func TestInvalidSocket(t *testing.T) {
	tests := []struct {
		name string
		in   []byte
	}{
		{
			"UnknownType",
			[]byte(`{"socket":{"other":{}}`),
		},
		{
			"StdIOParam",
			[]byte(`{"socket":{"stdio":{"param":0}}`),
		},
		{
			"WebsocketParam",
			[]byte(`{"socket":{"websocket":{"param":0}}`),
		},
		{
			"WebsocketCapRange",
			[]byte(`{"socket":{"websocket":{"max-connections":-1}}`),
		},
		{
			"WebsocketPortRange",
			[]byte(`{"socket":{"websocket":{"port":-1}}`),
		},
		{
			"SecureWebsocketParam",
			[]byte(`{"socket":{"secure-websocket":{"param":0}}`),
		},
		{
			"SecureWebsocketCapRange",
			[]byte(`{"socket":{"secure-websocket":{"max-connections":-1}}`),
		},
		{
			"SecureWebsocketPortRange",
			[]byte(`{"socket":{"secure-websocket":{"port":-1}}`),
		},
	}

	for _, test := range tests {
		t.Run(test.name, func(t *testing.T) {
			s := &Settings{}
			if err := s.parse(test.in); err == nil {
				t.Error("parse() did not fail")
			}
		})
	}
}

// TODO tests for new config options
