package jani

import (
	"dgit.cs.uni-saarland.de/pseuco/pseuco-jani-server/config"
	"errors"
	"github.com/gorilla/websocket"
	"io"
	"log"
	"sync"
	"time"
)

const (
	ProtocolVersion = 1

	Name        = "pseuCo JANI server"
	Author      = "Pascal Leo Held"
	Description = "A server-side implementation of the jani-interaction protocol written in Go."
	URL         = "https://dgit.cs.uni-saarland.de/pseuco/pseuco-jani-server"
)

var (
	cfg *config.Settings
	//connections      map[string]*wsConnection
	//connectionsGuard *sync.RWMutex
)

//func init() {
//	connections = make(map[string]*wsConnection)
//	connectionsGuard = &sync.RWMutex{}
//}

func logger(msg, when string) string {
	log.Println(when, msg)
	return msg
}

// Handler provides all facilities to serve incoming jani-interaction connections.
type Handler interface {
	// Handle manages a single jani-interaction connection.
	Handle()
}

type readWriteCloseFinalizer interface {
	io.ReadWriter
	Close(code int) error
	Finalize() error

	sync.Locker
	GetTask(id uint) (chan bool, error)
	SetTask(id uint, chn chan bool)
	DeleteTask(id uint)

	IsSecure() bool

	CanStart() bool
	Start()
	Stop()
	Timeout()
	GetKill() chan bool
}

type wsConnection struct {
	socket *websocket.Conn
	// FIXME
	//active      bool
	//msgIn       chan []byte
	//msgOutAsync chan []byte
	//msgOutSync  chan []byte
	//msgSent     chan bool
	//termReader  chan bool
	tasks      map[uint]chan bool
	tasksGuard *sync.Mutex
	reader     io.Reader
	closer     io.WriteCloser

	//ip          string
	secure      bool
	points      uint
	running     bool
	lastRunning time.Time

	canStart chan bool
	start    chan bool
	stop     chan bool
	timeout  chan bool
	kill     chan bool
}

// FIXME
type stdioConnection struct {
}

type opcodeError struct {
	Str string
}

func SetConfig(settings *config.Settings) {
	cfg = settings
}

// NewConnection constructs and returns a new Handler.
func NewConnection(socket *websocket.Conn, secure bool) Handler {
	//defer logger(logger("NewConnection", "enter:"), "leave:")
	//host, _, err := net.SplitHostPort(socket.UnderlyingConn().RemoteAddr().String())
	//if err != nil {
	//	log.Println("connect:", err)
	//}

	c := &wsConnection{
		socket: socket,
		// FIXME
		//active:      true,
		//msgIn:       make(chan []byte),
		//msgOutAsync: make(chan []byte),
		//msgOutSync:  make(chan []byte),
		//msgSent:     make(chan bool),
		//termReader:  make(chan bool, 1),
		tasks:      make(map[uint]chan bool),
		tasksGuard: &sync.Mutex{},
		//ip:         host,
		secure: secure,

		canStart: make(chan bool),
		start:    make(chan bool),
		stop:     make(chan bool),
		timeout:  make(chan bool),
		kill:     make(chan bool, 1),
	}

	//term := make(chan bool, 1)
	//go c.reader(term)
	//go c.writer(term)

	return c
}

func (c *wsConnection) Handle() {
	//defer logger(logger("Handle", "enter:"), "leave:")
	//connectionsGuard.Lock()
	//connections[c.ip] = c
	//connectionsGuard.Unlock()
	//defer func() {
	//	connectionsGuard.Lock()
	//	delete(connections, c.ip)
	//	connectionsGuard.Unlock()
	//}()

	term := make(chan bool)
	go c.watch(term)

	handle(c)

	term <- true
}

func (c *wsConnection) watch(term chan bool) {
	//defer logger(logger("watch", "enter:"), "leave:")
	var limits *config.RateLimit
	if c.secure {
		limits = cfg.Socket.SecureWebsocket.RateLimit
	} else {
		limits = cfg.Socket.Websocket.RateLimit
	}
	max := uint(float64(limits.Points.Max) * limits.Points.Grace)

	period := time.NewTicker(limits.Period)
	defer period.Stop()
	tick := time.NewTicker(limits.Tick)
	defer tick.Stop()
	for {
		cs := c.points < limits.Points.Max
		select {
		case <-term:
			return
		case c.canStart <- cs:
		case <-c.start:
			c.running = true
		case <-c.stop:
			c.running = false
			c.lastRunning = time.Now()
		case <-c.timeout:
			c.points += limits.Points.Timeout
		case <-period.C:
			c.points = 0
		case <-tick.C:
			if c.running {
				c.points += limits.Points.Tick

				if c.points > max {
					c.kill <- true
				}

				continue
			}

			if time.Now().Sub(c.lastRunning) > limits.Transition {
				c.points = saveSub(c.points, limits.Points.Tick)
			}
		}
	}
}

func saveSub(a, b uint) uint {
	if b > a {
		return 0
	}
	return a - b
}

func (c *wsConnection) Read(p []byte) (n int, err error) {
	//defer logger(logger("conn.Read", "enter:"), "leave:")
	if c.reader != nil {
		if n, err = c.reader.Read(p); err != nil {
			c.reader = nil
		} else {
			return
		}
	}

	var msgType int
	msgType, c.reader, err = c.socket.NextReader()
	if err != nil {
		return
	}

	if msgType != websocket.TextMessage {
		err = newOpcodeError()
		return
	}

	return c.reader.Read(p)
}

func (c *wsConnection) Write(p []byte) (n int, err error) {
	//defer logger(logger("conn.Write", "enter:"), "leave:")
	if c.closer != nil {
		if n, err = c.closer.Write(p); err != nil {
			_ = c.closer.Close()
			c.closer = nil
		} else {
			return
		}
	}

	c.closer, err = c.socket.NextWriter(websocket.TextMessage)
	if err != nil {
		return
	}

	n, err = c.closer.Write(p)
	if err != nil {
		return
	}

	err = c.closer.Close()
	if err != nil {
		return
	}

	return
}

func (c *wsConnection) Close(code int) (err error) {
	//defer logger(logger("conn.Close", "enter:"), "leave:")
	//c.active = false
	err = c.socket.WriteControl(
		websocket.CloseMessage,
		websocket.FormatCloseMessage(code, ""),
		time.Now().Add(time.Second),
	)
	return
}

func (c *wsConnection) Finalize() (err error) {
	//defer logger(logger("conn.Finalize", "enter:"), "leave:")
	for {
		if _, _, err = c.socket.NextReader(); err != nil {
			break
		}
	}
	if websocket.IsCloseError(err, websocket.CloseNoStatusReceived) {
		err = nil
	}
	return
}

func (c *wsConnection) Lock() {
	//defer logger(logger("conn.Lock", "enter:"), "leave:")
	c.tasksGuard.Lock()
}

func (c *wsConnection) Unlock() {
	//defer logger(logger("conn.Unlock", "enter:"), "leave:")
	c.tasksGuard.Unlock()
}

func (c *wsConnection) GetTask(id uint) (chan bool, error) {
	//defer logger(logger("conn.GetTask", "enter:"), "leave:")
	if chn, ok := c.tasks[id]; ok {
		return chn, nil
	} else {
		return nil, errors.New("not found")
	}
}

func (c *wsConnection) SetTask(id uint, chn chan bool) {
	//defer logger(logger("conn.SetTask", "enter:"), "leave:")
	c.tasks[id] = chn
}

func (c *wsConnection) DeleteTask(id uint) {
	//defer logger(logger("conn.DeleteTask", "enter:"), "leave:")
	delete(c.tasks, id)
}

func (c *wsConnection) IsSecure() bool {
	return c.secure
}

func (c *wsConnection) CanStart() bool {
	return <-c.canStart
}

func (c *wsConnection) Start() {
	c.start <- true
}

func (c *wsConnection) Stop() {
	c.stop <- true
}

func (c *wsConnection) Timeout() {
	c.timeout <- true
}

func (c *wsConnection) GetKill() chan bool {
	return c.kill
}

func (e *opcodeError) Error() string {
	return e.Str
}

func newOpcodeError() *opcodeError {
	return &opcodeError{Str: "jani-interaction protocol only supports WebSocket text messages"}
}

func isOpcodeError(err error) bool {
	if _, ok := err.(*opcodeError); ok {
		return true
	}
	return false
}
