/*
protocol state machine

	CONNECTED -Authenticate-> AUTHENTICATING -capabilitiesMsg-> INTERACTIVE
	                                |                             |
	                              Close                         Close
	                                |                             |
	                                v                             v
	                              CLOSED                        CLOSED
*/
package jani

import (
	"context"
	"encoding/json"
	"github.com/gorilla/websocket"
	"io"
	"log"
	"os/exec"
	"strings"
	"time"
)

func handle(connection readWriteCloseFinalizer) {
	//defer logger(logger("handle", "enter:"), "leave:")
	defer func() {
		if err := connection.Finalize(); err != nil {
			log.Println(err)
		}
	}()

	dec := json.NewDecoder(connection)
	enc := json.NewEncoder(connection)

	// CONNECTED -Authenticate-> AUTHENTICATING
	var auth authenticateMsg
	if err := dec.Decode(&auth); err != nil {
		log.Println("cannot decode authenticate message:", err)
		if err := connection.Close(websocket.CloseUnsupportedData); err != nil {
			log.Println("cannot close connection:", err)
		}
		return
	}

	// AUTHENTICATING -Close-> CLOSED
	if !checkVersion(&auth) {
		if err := enc.Encode(newCloseMsg(ProtocolVersionMismatch)); err != nil {
			log.Println("cannot encode close message:", err)
		}
		return
	}

	// AUTHENTICATING -Capabilities-> INTERACTIVE
	if err := enc.Encode(newCapabilitiesMsg()); err != nil {
		log.Println("cannot encode capabilities message:", err)
		if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
			log.Println("cannot close connection:", err)
		}
		return
	}

	for dec.More() {
		var message message
		if err := dec.Decode(&message); err != nil {
			log.Println("cannot decode message:", err)
			if isMessageTypeError(err) {
				if err := enc.Encode(newCloseMsg(NoTypeField)); err != nil {
					log.Println("cannot encode close message:", err)
				}
			}
			return
		}

		message.Handle(connection)
	}
}

func checkVersion(auth *authenticateMsg) bool {
	for _, version := range auth.JaniVersions {
		if version == ProtocolVersion {
			return true
		}
	}
	return false
}

func (m *message) Handle(connection readWriteCloseFinalizer) {
	//defer logger(logger("message.Handle", "enter:"), "leave:")
	m.handler.Handle(connection)
}

func (m *closeMsg) Handle(connection readWriteCloseFinalizer) {
	//defer logger(logger("close.Handle", "enter:"), "leave:")
	if err := connection.Close(websocket.CloseNormalClosure); err != nil {
		log.Println("cannot close connection:", err)
	}
}

func (m *serverParametersRequest) Handle(connection readWriteCloseFinalizer) {
	//defer logger(logger("serverParam.Handle", "enter:"), "leave:")
	if err := connection.Close(websocket.CloseNormalClosure); err != nil {
		log.Println("cannot close connection:", err)
	}
}

func (m *taskStop) Handle(connection readWriteCloseFinalizer) {
	//defer logger(logger("taskStop.Handle", "enter:"), "leave:")
	connection.Lock()
	if chn, err := connection.GetTask(m.Id); err == nil {
		chn <- true
	}
	connection.Unlock()
}

func (m *analysisEnginesRequest) Handle(connection readWriteCloseFinalizer) {
	//defer logger(logger("analysisEngineRequest.Handle", "enter:"), "leave:")
	enc := json.NewEncoder(connection)

	if err := enc.Encode(newAnalysisEnginesReply(m.Id)); err != nil {
		log.Println("cannot encode analysis engines message:", err)
		if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
			log.Println("cannot close connection:", err)
		}
	}

	return
}

func (m *analysisStart) Handle(connection readWriteCloseFinalizer) {
	//defer logger(logger("analysisStart.Handle", "enter:"), "leave:")
	enc := json.NewEncoder(connection)

	engine, ok := cfg.Engines.Engines[m.Engine]
	if !ok || m.ModellingFormalism != "x-pseuco" {
		if err := enc.Encode(newCloseMsg(MalformedStartRequest)); err != nil {
			log.Println("cannot encode close message:", err)
		}
		return
	}

	if !connection.CanStart() {
		if err := enc.Encode(newTaskMessage(m.Id, sevInfo, "submission limit reached")); err != nil {
			log.Println("cannot encode task status message:", err)
			if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
				log.Println("cannot close connection:", err)
			}
		}
		if err := enc.Encode(newTaskEnd(m.Id)); err != nil {
			log.Println("cannot encode task status message:", err)
			if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
				log.Println("cannot close connection:", err)
			}
		}
		return
	}

	term := make(chan bool, 1)
	connection.Lock()
	connection.SetTask(m.Id, term)
	connection.Unlock()

	connection.Start()
	defer connection.Stop()

	if err := enc.Encode(newTaskStatus(m.Id)); err != nil {
		log.Println("cannot encode task status message:", err)
		if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
			log.Println("cannot close connection:", err)
		}
	}

	var timeout time.Duration
	if connection.IsSecure() {
		timeout = cfg.Socket.SecureWebsocket.RateLimit.Timeout
	} else {
		timeout = cfg.Socket.Websocket.RateLimit.Timeout
	}
	kill := connection.GetKill()
	ctx, cancel := context.WithTimeout(context.Background(), timeout)
	defer cancel()
	go func() {
		for {
			select {
			case <-ctx.Done():
				return
			case <-kill:
				cancel()
			}
		}
	}()

	cmd := exec.CommandContext(ctx, *engine.Command, engine.Arguments...)
	stdin, err := cmd.StdinPipe()
	if err != nil {
		log.Println(err)
		if err := enc.Encode(newCloseMsg("cannot open stdin pipe")); err != nil {
			log.Println("cannot encode close message:", err)
		}
		return
	}

	go func() {
		//noinspection GoUnhandledErrorResult
		defer stdin.Close()
		_, _ = io.WriteString(stdin, m.Model.Parts[0].Part)
	}()

	output, err := cmd.CombinedOutput()
	if ctx.Err() == context.DeadlineExceeded {
		connection.Timeout()
		log.Println("Command timed out!")
		if err := enc.Encode(newTaskMessage(m.Id, sevError, "Execution exceeded time limit!")); err != nil {
			log.Println("cannot encode task status message:", err)
			if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
				log.Println("cannot close connection:", err)
			}
		}
		if err := enc.Encode(newTaskEnd(m.Id)); err != nil {
			log.Println("cannot encode task status message:", err)
			if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
				log.Println("cannot close connection:", err)
			}
		}
		return
	}

	if err != nil {
		log.Println(err)
		log.Println(string(output))
		lines := strings.Split(string(output), "\n")
		var modified []string
		modified = append(modified, "The output was:\n")
		for _, line := range lines {
			if strings.HasPrefix(line, "Note: No input file specified. Use standard input:") {
				continue
			}
			line = strings.Replace(line, "<stdin>", m.Model.Parts[0].Name, 1)
			modified = append(modified, line)
		}
		if err := enc.Encode(newTaskMessage(m.Id, sevError, "Execution of "+*engine.Name+" terminated prematurely!")); err != nil {
			log.Println("cannot encode task status message:", err)
			if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
				log.Println("cannot close connection:", err)
			}
		}
		if err := enc.Encode(newTaskMessage(m.Id, sevInfo, strings.Join(modified, "\n"))); err != nil {
			log.Println("cannot encode task status message:", err)
			if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
				log.Println("cannot close connection:", err)
			}
		}
		if err := enc.Encode(newTaskEnd(m.Id)); err != nil {
			log.Println("cannot encode task status message:", err)
			if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
				log.Println("cannot close connection:", err)
			}
		}
		return
	}

	lines := strings.Split(string(output), "\n")
	severity := sevInfo
	for _, line := range lines {
		if strings.HasPrefix(line, "Note:") || strings.HasPrefix(line, "Total:") {
			if strings.HasPrefix(line, "Note: No input file specified. Use standard input:") {
				continue
			}
			severity = sevInfo
		}
		if strings.HasPrefix(line, "Warning:") {
			severity = sevWarning
		}
		if strings.HasPrefix(line, "Error:") {
			severity = sevError
		}
		if err := enc.Encode(newTaskMessage(m.Id, severity, line)); err != nil {
			log.Println("cannot encode task status message:", err)
			if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
				log.Println("cannot close connection:", err)
			}
		}
	}

	connection.Lock()
	connection.DeleteTask(m.Id)
	connection.Unlock()

	if err := enc.Encode(newTaskEnd(m.Id)); err != nil {
		log.Println("cannot encode task status message:", err)
		if err := connection.Close(websocket.CloseInternalServerErr); err != nil {
			log.Println("cannot close connection:", err)
		}
	}
}
