package jani

import (
	"encoding/json"
)

// reasons for closing connection
const (
	ProtocolVersionMismatch = "jani-interaction protocol version mismatch"
	NoTypeField             = "message has no type field"
	TypeFieldNotString      = "type field is not of type string"
	UnknownType             = "type unknown"
	MalformedStartRequest   = "malformed analysis start request"
)

// available message types
const (
	// general
	typeClose = "close"

	// server
	typeCapabilities     = "capabilities"
	typeServerParameters = "server-parameters"

	// tasks
	typeTaskStatus   = "task-status"
	typeTaskProgress = "task-progress"
	typeTaskMessage  = "task-message"
	typeTaskStop     = "task-stop"
	typeTaskEnd      = "task-end"

	// analysis
	typeAnalysisEngines = "analysis-engines"
	typeAnalysisStart   = "analysis-start"
	typeAnalysisResults = "analysis-results"
)

type severity string

const (
	sevInfo    severity = "info"
	sevWarning severity = "warning"
	sevError   severity = "error"
)

type messageTypeError struct {
	Str string
}

/*
 * MESSAGES
 */

type messageHandler interface {
	Handle(connection readWriteCloseFinalizer)
}

type message struct {
	handler messageHandler
}

// CONNECTED -Authenticate-> AUTHENTICATING
type authenticateMsg struct {
	JaniVersions []uint `json:"jani-versions"`

	// currently not supported
	Extensions []string `json:"extensions,omitempty"`
	Login      string   `json:"login,omitempty"`
	Password   string   `json:"password,omitempty"`
}

// AUTHENTICATING & INTERACTIVE
type closeMsg struct {
	Type   string `json:"type"`
	Reason string `json:"reason,omitempty"`
}

// AUTHENTICATING -Capabilities-> INTERACTIVE
type capabilitiesMsg struct {
	Type        string   `json:"type"`
	JaniVersion uint     `json:"jani-version"`
	Metadata    metadata `json:"metadata"`
	Roles       []string `json:"roles"`

	// currently not supported
	Extensions []string `json:"extensions,omitempty"`
	// TODO parameters
}

// INTERACTIVE

// server parameters (currently not supported)
type serverParametersRequest struct {
	Type       string           `json:"type"`
	Id         uint             `json:"id"`
	Parameters []parameterValue `json:"parameters"`
}

type serverParametersReply struct {
	Type    string `json:"type"`
	Id      uint   `json:"id"`
	Success bool   `json:"success"`
	Error   string `json:"error,omitempty"`
}

// tasks
type taskStatus struct {
	Type   string `json:"type"`
	Id     uint   `json:"id"`
	Status string `json:"status"`
}

type taskProgress struct {
	Type     string  `json:"type"`
	Id       uint    `json:"id"`
	Progress float64 `json:"progress"`
}

type taskMessage struct {
	Type     string   `json:"type"`
	Id       uint     `json:"id"`
	Severity severity `json:"severity"`
	Message  string   `json:"message"`
}

type taskStop struct {
	Type string `json:"type"`
	Id   uint   `json:"id"`
}

type taskEnd struct {
	Type string `json:"type"`
	Id   uint   `json:"id"`
}

// analysis
type analysisEnginesRequest struct {
	Type string `json:"type"`
	Id   uint   `json:"id"`
}

type analysisEnginesReply struct {
	Type    string   `json:"type"`
	Id      uint     `json:"id"`
	Engines []engine `json:"engines"`
}

type analysisStart struct {
	Type               string   `json:"type"`
	Id                 uint     `json:"id"`
	Engine             string   `json:"engine"`
	ModellingFormalism string   `json:"modelling-formalism,omitempty"`
	Model              anyModel `json:"model"`

	// currently not supported
	Properties  []string         `json:"properties,omitempty"`
	Parameters  []parameterValue `json:"parameters"`
	Experiments []experiment     `json:"experiments,omitempty"`
}

type analysisResults struct {
	Type    string          `json:"type"`
	Id      uint            `json:"id"`
	Results analysisDataSet `json:"results"`
}

/*
 * UTIL / SUPPORT
 */

type metadata struct {
	Name        string  `json:"name"`
	Version     version `json:"version"`
	Author      string  `json:"author,omitempty"`
	Description string  `json:"description,omitempty"`
	Url         string  `json:"url,omitempty"`
}

type version struct {
	Major    uint `json:"major"`
	Minor    uint `json:"minor"`
	Revision uint `json:"revision"`
}

type engine struct {
	Id       string   `json:"id"`
	Metadata metadata `json:"metadata"`

	// currently not supported
	// TODO parameters
	ModellingFormalisms []string `json:"modelling-formalisms,omitempty"`
}

type anyModel struct {
	Name  string `json:"name"`
	Parts []part `json:"parts"`
}

type part struct {
	Name string `json:"name"`
	Role string `json:"role"`
	Part string `json:"part"`
}

type experiment struct {
	Experiment string  `json:"experiment"`
	Values     []value `json:"values"`
}

type value struct {
	Name  string      `json:"name"`
	Value interface{} `json:"value"`
}

type analysisDataSet struct {
	Id         uint              `json:"id"`
	Label      string            `json:"label,omitempty"`
	Experiment string            `json:"experiment,omitempty"`
	Property   string            `json:"property,omitempty"`
	Results    []result          `json:"results,omitempty"`
	DataSets   []analysisDataSet `json:"data-sets"`
}

type result struct {
	Label          string      `json:"label"`
	Type           string      `json:"type"`
	Value          interface{} `json:"value"`
	Unit           string      `json:"unit,omitempty"`
	FormattedValue string      `json:"formatted-value,omitempty"`
}

// currently not supported
type parameterValue struct {
	Id    string      `json:"id"`
	Value interface{} `json:"value"`
}

/*
 * Methods
 */

func (e *messageTypeError) Error() string {
	return e.Str
}

func newMessageTypeError(str string) *opcodeError {
	return &opcodeError{Str: str}
}

func isMessageTypeError(err error) bool {
	if _, ok := err.(*messageTypeError); ok {
		return true
	}
	return false
}

func (m *message) UnmarshalJSON(data []byte) (err error) {
	var content map[string]interface{}
	if err = json.Unmarshal(data, &content); err != nil {
		return
	}

	mt, ok := content["type"]
	if !ok {
		err = newMessageTypeError(NoTypeField)
		return
	}

	var msgType string
	if msgType, ok = mt.(string); !ok {
		err = newMessageTypeError(TypeFieldNotString)
		return
	}

	switch msgType {
	case typeClose:
		var msg closeMsg
		if err = json.Unmarshal(data, &msg); err != nil {
			return
		}
		m.handler = &msg
	case typeServerParameters:
		var msg serverParametersRequest
		if err = json.Unmarshal(data, &msg); err != nil {
			return
		}
		m.handler = &msg
	case typeTaskStop:
		var msg taskStop
		if err = json.Unmarshal(data, &msg); err != nil {
			return
		}
		m.handler = &msg
	case typeAnalysisEngines:
		var msg analysisEnginesRequest
		if err = json.Unmarshal(data, &msg); err != nil {
			return
		}
		m.handler = &msg
	case typeAnalysisStart:
		var msg analysisStart
		if err = json.Unmarshal(data, &msg); err != nil {
			return
		}
		m.handler = &msg
	default:
		err = newMessageTypeError(UnknownType)
	}

	return
}

func newCloseMsg(reason string) *closeMsg {
	return &closeMsg{
		Type:   typeClose,
		Reason: reason,
	}
}

func newCapabilitiesMsg() *capabilitiesMsg {
	return &capabilitiesMsg{
		Type:        typeCapabilities,
		JaniVersion: ProtocolVersion,
		Metadata: metadata{
			Name: Name,
			Version: version{
				Major:    cfg.Version.Major,
				Minor:    cfg.Version.Minor,
				Revision: cfg.Version.Patch,
			},
			Author:      Author,
			Description: Description,
			Url:         URL,
		},
		Roles: []string{"analyse"}, // "transform" role currently not supported
	}
}

func newAnalysisEnginesReply(request uint) *analysisEnginesReply {
	msg := &analysisEnginesReply{
		Type:    typeAnalysisEngines,
		Id:      request,
		Engines: []engine{},
	}

	for id, e := range cfg.Engines.Engines {
		engine := engine{
			Id: id,
			Metadata: metadata{
				Name: *e.Name,
				Version: version{
					Major:    e.Version.Major,
					Minor:    e.Version.Minor,
					Revision: e.Version.Patch,
				},
				Author:      e.Author,
				Description: e.Description,
				Url:         e.Url,
			},
			ModellingFormalisms: []string{"x-pseuco"},
		}
		msg.Engines = append(msg.Engines, engine)
	}

	return msg
}

func newTaskStatus(request uint) *taskStatus {
	return &taskStatus{
		Type:   typeTaskStatus,
		Id:     request,
		Status: "Running",
	}
}

func newTaskMessage(request uint, severity severity, msg string) *taskMessage {
	return &taskMessage{
		Type:     typeTaskMessage,
		Id:       request,
		Severity: severity,
		Message:  msg,
	}
}

func newTaskEnd(request uint) *taskEnd {
	return &taskEnd{
		Type: typeTaskEnd,
		Id:   request,
	}
}
