/*
jani-interaction server implementation
Accepts connections on standard streams and websocket (in/-secure).
*/

package main

import (
	"dgit.cs.uni-saarland.de/pseuco/pseuco-jani-server/config"
	"dgit.cs.uni-saarland.de/pseuco/pseuco-jani-server/jani"
	"flag"
	"fmt"
	"github.com/gorilla/websocket"
	"log"
	"net/http"
	"os"
	"regexp"
)

var cfgFile string

func init() {
	const (
		defaultConfig = "config.json"
		usage         = "the configuration file"
	)
	flag.StringVar(&cfgFile, "config", defaultConfig, usage)
	flag.StringVar(&cfgFile, "c", defaultConfig, usage+" (shorthand)")
}

var version bool

func init() {
	const (
		usage = "print version and quit"
	)
	flag.BoolVar(&version, "version", false, usage)
	flag.BoolVar(&version, "v", false, usage+" (shorthand)")
}

var stdout bool

func init() {
	flag.BoolVar(&stdout, "stdout", false, "use stdout for logging")
}

func main() {
	flag.Parse()

	cfg, err := config.Read(cfgFile)
	if err != nil {
		fmt.Println(err)
		os.Exit(1)
	}

	if version {
		fmt.Println("pseuco-jani-server version:", cfg.Version)
		return
	}

	if stdout && cfg.Socket.StdIO == nil {
		log.SetOutput(os.Stdout)
	} else {
		f, _ := os.OpenFile("server.log", os.O_RDWR|os.O_CREATE|os.O_APPEND, 0666)
		// FIXME
		//f, _ := os.OpenFile("server.log", os.O_RDWR|os.O_CREATE|os.O_TRUNC, 0666)
		//noinspection GoUnhandledErrorResult
		defer f.Close()
		log.SetOutput(f)
	}
	log.Println("server started")

	jani.SetConfig(cfg)

	initConnections(cfg.Socket)
	initHttpHandler()

	listenAndServe(cfg)
}

var (
	insecureConnections chan bool
	secureConnections   chan bool
)

func initConnections(socket *config.Socket) {
	if socket.Websocket != nil && socket.Websocket.MaxConnections > 0 {
		insecureConnections = make(chan bool, socket.Websocket.MaxConnections)
	}
	if socket.SecureWebsocket != nil && socket.SecureWebsocket.MaxConnections > 0 {
		secureConnections = make(chan bool, socket.SecureWebsocket.MaxConnections)
	}
}

func initHttpHandler() {
	http.HandleFunc("/", handleRoot)
}

func handleRoot(w http.ResponseWriter, r *http.Request) {
	if !websocket.IsWebSocketUpgrade(r) {
		return
	}

	var connections chan bool
	secure := r.TLS != nil
	if !secure {
		connections = insecureConnections
	} else {
		connections = secureConnections
	}

	if connections != nil {
		select {
		case connections <- true:
		default:
			return
		}
	}

	wsConn, err := upgrader.Upgrade(w, r, nil)
	if err != nil {
		log.Println(err)
		return
	}
	defer func() {
		if err := wsConn.Close(); err != nil {
			log.Println("cannot close WebSocket connection:", err)
		}
	}()

	janiConn := jani.NewConnection(wsConn, secure)
	janiConn.Handle()

	if connections != nil {
		<-connections
	}
}

var upgrader = websocket.Upgrader{
	CheckOrigin: checkOrigin,
}

func checkOrigin(r *http.Request) bool {
	return originRegex.MatchString(r.Header.Get("Origin"))
}

var originRegex = regexp.MustCompile(`^https?://((.+\.)?pseuco\.com|(.+\.)?localhost|127\.0\.0\.1)`)

func listenAndServe(cfg *config.Settings) {
	ioTerm := make(chan bool, 1)
	if cfg.Socket.StdIO != nil {
		go func() {
			// FIXME
			conn := jani.NewConnection(nil, false)
			conn.Handle()
			ioTerm <- true
		}()
	} else {
		ioTerm <- true
	}

	wsTerm := make(chan bool, 1)
	if cfg.Socket.Websocket != nil {
		go func() {
			port := cfg.Socket.Websocket.Port
			if err := http.ListenAndServe(fmt.Sprintf(":%d", port), nil); err != nil {
				log.Println(err)
			}
			wsTerm <- true
		}()
	} else {
		wsTerm <- true
	}

	swsTerm := make(chan bool, 1)
	if cfg.Socket.SecureWebsocket != nil {
		go func() {
			port := cfg.Socket.SecureWebsocket.Port
			if err := http.ListenAndServeTLS(
				fmt.Sprintf(":%d", port),
				cfg.Socket.SecureWebsocket.Certificate,
				cfg.Socket.SecureWebsocket.PrivateKey,
				nil); err != nil {
				log.Println(err)
			}
			swsTerm <- true
		}()
	} else {
		swsTerm <- true
	}

	<-ioTerm
	<-wsTerm
	<-swsTerm
}
